from abc import ABC
import os
from typing import Optional, Any
from importlib.metadata import version
from decimal import Decimal
from urllib.parse import urlparse
import fsspec

class BaseEngine(ABC):
    """
    Abstract base class for implementing different engine types.

    Attributes
    ----------
    SQLGLOT_DIALECT : str, optional
        Specifies the SQL dialect to be used by the engine when SQL transpiling
        is required. Default is None.
    SUPPORTS_ONELAKE : bool
        Indicates if the engine supports OneLake URIs (e.g., abfss://workspace@onelake.dfs.fabric.microsoft.com/...)
    SUPPORTS_SCHEMA_PREP : bool
        Indicates if the engine supports schema preparation (creation of empty table with defined schema)
    SUPPORTS_MOUNT_PATH : bool
        Indicates if the engine supports mount URIs (e.g., /mnt/...)

    Methods
    -------
    get_total_cores()
        Returns the total number of CPU cores available on the system.
    get_compute_size()
        Returns a formatted string with the compute size.
    append_array_to_delta(abfss_path: str, array: list)
        Appends a list of data to a Delta table at the specified path.
    """
    SQLGLOT_DIALECT = None
    SUPPORTS_SCHEMA_PREP = False
    SUPPORTS_MOUNT_PATH = True
    TABLE_FORMAT = 'delta'
    
    def __init__(
            self, 
            schema_or_working_directory_uri: str = None,
            storage_options: Optional[dict[str, Any]] = None
            ):
        """
        Parameters
        ----------
        schema_or_working_directory_uri : str, optional
            The base URI where tables are stored. For non-Spark engines, 
            tables are stored directly under this path. For Spark engines, 
            this serves as the root schema path where tables are created.
        storage_options : dict, optional
            A dictionary of storage options to pass to the engine for filesystem access.
        """
        self.version: str = ''
        self.cost_per_vcore_hour: Optional[float] = None
        self.cost_per_hour: Optional[float] = None
        self.extended_engine_metadata: dict[str, str] = {}
        self.storage_options: dict[str, Any] = storage_options if storage_options is not None else {}
        self.schema_or_working_directory_uri: str = schema_or_working_directory_uri.replace("file:///", "").replace(chr(92), '/') if schema_or_working_directory_uri else None

        self.runtime = self._detect_runtime() if getattr(self, 'runtime', None) is None else self.runtime
        self.operating_system = self._detect_os() if getattr(self, 'operating_system', None) is None else self.operating_system

        if self.runtime == "fabric":
            from IPython.core.getipython import get_ipython
            import sempy.fabric as fabric

            self._notebookutils = get_ipython().user_ns.get("notebookutils")
            self._fabric_rest = fabric.FabricRestClient()
            workspace_id = self._notebookutils.runtime.context['currentWorkspaceId']
            self.region = self._fabric_rest.get(path_or_url=f"/v1/workspaces/{workspace_id}").json()['capacityRegion'].replace(' ', '').lower()
            self.capacity_id = self._fabric_rest.get(path_or_url=f"/v1/workspaces/{workspace_id}").json()['capacityId']
            self._autocalc_usd_cost_per_vcore_hour = self._get_vm_retail_rate(self.region, 'Spark Memory Optimized Capacity Usage')
            self.extended_engine_metadata.update({'compute_region': self.region})
            # rust object store (used by delta-rs, polars, sail) parametrization; https://docs.rs/object_store/latest/object_store/azure/enum.AzureConfigKey.html#variant.Token
            os.environ["AZURE_STORAGE_TOKEN"] = self._notebookutils.credentials.getToken("storage")

        self.extended_engine_metadata.update({
            'runtime': self.runtime,
            'os': self.operating_system
        })

        if self.schema_or_working_directory_uri is None:
            self.fs = None
        elif self.runtime in ("fabric", "synapse"):
            # workaround: use notebookutils filesystem for abfs due to recursive delete issues in fsspec
            # https://github.com/developmentseed/obstore/issues/556
            self.fs = self._notebookutils.fs
            self.fs.mkdir = self.fs.mkdirs # notebookutils users mkdirs
            if self.storage_options == {}:
                self._validate_and_set_azure_storage_config()
        elif urlparse(self.schema_or_working_directory_uri).scheme in ("s3", "gs"):
            # TODO: test s3 and gs support
            self.fs = fsspec.filesystem(urlparse(self.schema_or_working_directory_uri).scheme, **self.storage_options)
        else:
            # TODO: use FsspecStore once it's #555 and #556 are fixed
            # workaround: use original fsspec until obstore bugs are fixes:
            # * https://github.com/developmentseed/obstore/issues/555
            self.fs = fsspec.filesystem("file")

    def _detect_runtime(self) -> str:
        """
        Dynamically detect the runtime/environment.
        Returns: str - The detected service name
        """
        import os    

        # Check for Microsoft Fabric or Synapse
        try:
            notebookutils = None
            utils_modules = ('notebookutils', 'mssparkutils')
            for utils_module in utils_modules:
                try:
                    notebookutils = __import__(utils_module)
                except ImportError:
                    continue
            if notebookutils and hasattr(notebookutils, 'runtime'):
                if hasattr(notebookutils.runtime, 'context'):
                    context = notebookutils.runtime.context
                    if 'productType' in context:
                        product = context['productType'].lower()
                        return product
        except:
            pass
        
        # Check for Databricks
        try:
            dbutils = None
            if 'DATABRICKS_RUNTIME_VERSION' in os.environ:
                return "databricks"
            try:
                dbutils = __import__('dbutils')
                if dbutils is not None:
                    return "databricks"
            except:
                pass
        except:
            pass
        
        # Check for Google Colab
        try:
            if 'COLAB_RELEASE_TAG' in os.environ:
                return "colab"
        except ImportError:
            pass
        
        # Default fallback
        return "local_unknown"

    def _detect_os(self) -> str:
        """
        Dynamically detect the operating system.
        Returns: str - The detected OS name
        """
        import sys

        os_platform = sys.platform.lower()
        if os_platform.startswith('win'):
            return 'windows'
        elif os_platform.startswith('linux'):
            return 'linux'
        elif os_platform.startswith('darwin'):
            return 'mac'
        else:
            return 'unknown'

    def _validate_and_set_azure_storage_config(self) -> None:
        if not os.getenv("AZURE_STORAGE_TOKEN"):
            raise ValueError("""Please store bearer token as env variable `AZURE_STORAGE_TOKEN` (via `os.environ["AZURE_STORAGE_TOKEN"] = "..."`)""")
        self.storage_options = {
            "bearer_token": os.getenv("AZURE_STORAGE_TOKEN"),
            "allow_invalid_certificates": "true",  # https://github.com/delta-io/delta-rs/issues/3243#issuecomment-2727206866
        }

    def _get_vm_retail_rate(self, region: str, sku: str, spot: bool = False) -> float:
        import requests
        query = f"armRegionName eq '{region}' and serviceName eq 'Microsoft Fabric' and skuName eq '{sku}'"
        api_url = "https://prices.azure.com/api/retail/prices?"
        return requests.get(api_url, params={'$filter': query}).json()['Items'][0]['retailPrice'] / 2
    
    def get_total_cores(self) -> int:
        """
        Returns the total number of CPU cores available on the system.
        """
        cores = os.cpu_count()
        return cores
    
    def get_compute_size(self) -> str:
        """
        Returns a formatted string with the compute size.
        """
        cores = self.get_total_cores()
        return f"{cores}vCore"
    
    def get_job_cost(self, duration_ms: int) -> Optional[Decimal]:
        """
        Returns the cost per hour for compute as a Decimal.
        
        If `cost_per_vcore_hour` or `cost_per_hour` is provided, it calculates the job cost.
        Otherwise, it returns None.
        """
        if self.cost_per_hour is None and self.cost_per_vcore_hour is not None:
            self.cost_per_hour = Decimal(self.get_total_cores()) * Decimal(self.cost_per_vcore_hour)
        elif self.cost_per_hour is None:
            return None

        job_cost = Decimal(self.cost_per_hour) * (Decimal(duration_ms) / Decimal(3600000))  # Convert ms to hours
        return job_cost.quantize(Decimal('0.0000000000'))  # Ensure precision matches DECIMAL(18,10)
    
    
    def create_external_location(self, location_uri: str):
        """
        Supports engines that need to create external locations for data access.
        By default, this is a no-op and is only overridden by subclasses as needed.
        """
        pass
    
    def create_schema_if_not_exists(self, drop_before_create: bool = True):
        if drop_before_create:
            if self.fs.exists(self.schema_or_working_directory_uri):
                self.fs.rm(self.schema_or_working_directory_uri, True)
            self.fs.mkdir(self.schema_or_working_directory_uri)
    
    def _convert_generic_to_specific_schema(self, generic_schema: list):
        """
        Convert a generic schema to a specific Spark schema.
        """
        import pyarrow as pa
        type_mapping = {
            'STRING': pa.string(),
            'TIMESTAMP': pa.timestamp('us', tz='UTC'),
            'TINYINT': pa.int8(),
            'SMALLINT': pa.int16(),
            'INT': pa.int32(),
            'BIGINT': pa.int64(),
            'FLOAT': pa.float32(),
            'DOUBLE': pa.float64(),
            'DECIMAL(18,10)': pa.decimal128(18, 10),
            'BOOLEAN': pa.bool_(),
            'MAP<STRING, STRING>': pa.map_(pa.string(), pa.string())
        }
        return pa.schema([(name, type_mapping[data_type]) for name, data_type in generic_schema])
    
    def _append_results_to_delta(self, table_uri: str, results: list, generic_schema: list):
        """
        Appends a list of result records to an existing Delta table.

        Parameters
        ----------
        table_uri : str
            The URI of where the table resides.
        results : list of dict
            A list of row dictionaries to append. Each dictionary may include an
            'engine_metadata' key, whose contents will be stored as a MAP<STRING,STRING>.
        generic_schema : list of tuple
            A list of (field_name, field_type) tuples defining the generic schema
            for the rows in `results`.

        Notes
        -----
        - Converts `generic_schema` into a PyArrow schema.
        - Extracts 'engine_metadata' from each row and appends it as a
          MAP<STRING,STRING> column.
        - Uses DeltaRs to write the data in "append" mode.
        - If the installed `deltalake` version is 0.x, forces the Rust engine.
        """
        import pyarrow as pa
        from ..engines.delta_rs import DeltaRs

        schema = self._convert_generic_to_specific_schema(generic_schema=generic_schema)

        # Extract engine_metadata and convert to map format, otherwise is interpreted as a Struct
        index = schema.get_field_index("engine_properties")
        schema = schema.remove(index)
        index = schema.get_field_index("execution_telemetry")
        schema = schema.remove(index)

        engine_map_data = []
        execution_map_data = []
        for result in results:
            engine_properties = result.pop('engine_properties', {})
            if engine_properties:
                map_items = [(str(k), str(v)) for k, v in engine_properties.items()]
            else:
                map_items = []

            engine_map_data.append(map_items)

            execution_telemetry = result.pop('execution_telemetry', {})
            if execution_telemetry:
                execution_map_items = [(str(k), str(v)) for k, v in execution_telemetry.items()]
            else:
                execution_map_items = []

            execution_map_data.append(execution_map_items)

        table = pa.Table.from_pylist(results, schema)
        engine_map_array = pa.array(engine_map_data, type=pa.map_(pa.string(), pa.string()))
        execution_map_array = pa.array(execution_map_data, type=pa.map_(pa.string(), pa.string()))
        table = table.append_column('engine_properties', engine_map_array)
        table = table.append_column('execution_telemetry', execution_map_array)

        if version('deltalake').startswith('0.'):
            DeltaRs().write_deltalake(
                table_uri, 
                table, 
                mode="append",
                schema_mode='merge',
                engine='rust'
            )
        else:
            DeltaRs().write_deltalake(
                table_or_uri=table_uri,
                data=table,
                mode="append",
                schema_mode="merge",
                storage_options=self.storage_options,
            )


from ._tpc_rs import _TPCRsDataGenerator
class TPCHDataGenerator(_TPCRsDataGenerator):
    """
    This class is a multithreading wrapper of the rust-based TPC-H data generator, `tpchgen-rs`. It generates TPC-H data in Parquet format
    based on the specified scale factor, target row group size in MB, and compression codec. Each table is partitioned into multiple parts to
    target generating 1GB sized files.

    Attributes
    ----------
    scale_factor : int
        The scale factor for the data generation, which determines the size of the generated dataset.
    target_folder_uri : str, optional
        The folder path where the generated Parquet data will be stored. A folder for each table will be created.
    target_row_group_size_mb : int
        The target size of row groups in megabytes for the generated Parquet files.
    compression: str, default="ZSTD"
        Compression codec to use for the generated parquet files.
        Supports codecs: "UNCOMPRESSED", "SNAPPY", "GZIP(compression_level)", "BROTLI(compression_level)", "LZ4", "LZ4_RAW", "LZO", "ZSTD(compression_level)"

    Methods
    -------
    run()
        Generates TPC-H data in Parquet format based on the input scale factor and writes it to the target folder.
    """
    GEN_UTIL = 'dbgen'
    GEN_TYPE = 'tpch'
    GEN_SF1000_FILE_COUNT_MAP = {
        'lineitem': 150,
        'orders': 40,
        'partsupp': 26,
        'part': 4,
        'customer': 8
    }
    GEN_TABLE_REGISTRY = [
        'customer', 'lineitem', 'nation', 'orders', 'part',
        'partsupp', 'region', 'supplier'
    ]
    SF1000_SIZE_GB_DICT = {
        'lineitem':  152,
        'orders': 38,
        'partsupp': 26.7,
        'part': 4,
        'customer': 7.6,
        'supplier': 0.48,
        'region': 0.00,
        'nation': 0.00
    }


from .spark import Spark
from typing import Optional
from decimal import Decimal
import re
from urllib.parse import urlparse, parse_qs

class FabricSpark(Spark):
    """
    Fabric Spark Engine
    """

    def __init__(
            self,
            lakehouse_name: str, 
            lakehouse_schema_name: str,
            spark_measure_telemetry: bool = False,
            cost_per_vcore_hour: Optional[float] = None,
            compute_stats_all_cols: bool = False
            ):
        """
        Parameters
        ----------
        lakehouse_name : str
            The name of the lakehouse (catalog) to use within Fabric.
        lakehouse_schema_name : str
            The name of the schema (database) to use within the catalog.
        spark_measure_telemetry : bool, default False
            Whether to enable sparkmeasure telemetry for performance measurement.
        cost_per_vcore_hour : float, optional
            The cost per vCore hour for the Spark cluster. If None, cost calculations are auto calculated
            where possible.
        compute_stats_all_cols : bool, default False
            Whether to stats should be computed automatically as part of write operations.
        """

        super().__init__(
            catalog_name=lakehouse_name, 
            schema_name=lakehouse_schema_name, 
            spark_measure_telemetry=spark_measure_telemetry, 
            cost_per_vcore_hour=cost_per_vcore_hour,
            compute_stats_all_cols=compute_stats_all_cols
        )

        self.version: str = f"{self.spark.sparkContext.version} (vhd_name=={self.spark.conf.get('spark.synapse.vhd.name')})"
        self.cost_per_vcore_hour = cost_per_vcore_hour or getattr(self, '_autocalc_usd_cost_per_vcore_hour', None)
        self.cost_per_hour = self.get_total_cores() * self.cost_per_vcore_hour

        url = self.spark.sparkContext.uiWebUrl
        # Parse webUrl string
        parsed = urlparse(url)
        query = parse_qs(parsed.query)
        artifact_id = query.get("artifactId", [None])[0]
        # Regex for GUIDs
        guid_pattern = re.compile(r"[0-9a-fA-F-]{36}")
        guids = guid_pattern.findall(url)
        tenant_id = guids[0]     # after /sparkui/
        activity_id = guids[2]  # after /activities/

        self.extended_engine_metadata.update({
            'spark_history_url': f"https://{self.spark_configs['spark.trident.pbienv'].lower()}.powerbi.com/workloads/de-ds/sparkmonitor/{artifact_id}/{activity_id}?ctid={tenant_id}",
            'cost_per_hour': Decimal(self.cost_per_hour).quantize(Decimal('0.0000')),
            'capacity_id': self.capacity_id
        })

        spark_configs_to_log = {k: v for k, v in self.spark_configs.items() if k in [
            'spark.sql.parquet.vorder.enabled',
            'spark.sql.parquet.vorder.default',
            'spark.microsoft.delta.optimizeWrite.enabled',
            'spark.microsoft.delta.optimizeWrite.binSize',
            'spark.synapse.vegas.useCache',
            'spark.synapse.vegas.cacheSize',
            'spark.native.enabled',
            'spark.gluten.enabled',
            'spark.sql.parquet.native.writer.directWriteEnabled',
            'spark.synapse.vhd.name',
            'spark.synapse.vhd.id',
            'spark.microsoft.delta.stats.collect.extended',
            'spark.microsoft.delta.stats.injection.enabled',
            'spark.microsoft.delta.snapshot.driverMode.enabled',
            'spark.microsoft.delta.stats.collect.extended.property.setAtTableCreation',
            'spark.microsoft.delta.targetFileSize.adaptive.enabled',
            'spark.app.id',
            'spark.cluster.name'
        ]}

        self.extended_engine_metadata.update(spark_configs_to_log)

        self.compute_stats_all_cols = compute_stats_all_cols
        self.run_analyze_after_load = False # Fabric Spark supports auto stats collection
        if self.compute_stats_all_cols:
            # Enable auto stats collection
            self.spark.conf.set("spark.microsoft.delta.stats.collect.extended", "true")
            self.spark.conf.set("spark.microsoft.delta.stats.injection.enabled", "true")
            self.spark.conf.set("spark.microsoft.delta.stats.collect.extended.property.setAtTableCreation", "true")



import posixpath
import importlib.util
import fsspec
from fsspec import AbstractFileSystem
from lakebench.utils.path_utils import to_unix_path

class _TPCDataGenerator:
    """
    Base class for TPC data generation. PLEASE DO NOT INSTANTIATE THIS CLASS DIRECTLY. Use the TPCHDataGenerator and TPCDSDataGenerator
    subclasses instead.
    """
    GEN_UTIL = ''
    GEN_TYPE = ''

    def __init__(self, scale_factor: int, target_folder_uri: str, target_row_group_size_mb: int = 128) -> None:
        """
        Initialize the TPC data generator with a scale factor.

        Parameters
        ----------
        scale_factor: int
            The scale factor for the data generation.
        target_folder_uri: str
            Test data will be written to this location where tables are represented as folders containing parquet files.
        target_row_group_size_mb: int, default=128
            Desired row group size for the generated parquet files.

        """
        self.scale_factor = scale_factor
        if target_folder_uri.startswith("abfss://"):
            raise ValueError("abfss path currently not supported. DuckDB is used for data generation and DuckDB is not able to write to Azure remote storage as of now.")
            # self.fs: FsspecStore = FsspecStore(protocol=urlparse(target_mount_folder_path).scheme)
        else:
            # workaround: use original fsspec until obstore bugs are fixes:
            # * https://github.com/developmentseed/obstore/issues/555
            self.fs: AbstractFileSystem = fsspec.filesystem("file")
        self.target_folder_uri = to_unix_path(target_folder_uri)
        self.target_row_group_size_mb = target_row_group_size_mb

        if importlib.util.find_spec("duckdb") is None:
            raise ImportError(
                "DuckDB is used for data generation but is not installed. Install using `%pip install lakebench[duckdb]` or `%pip install lakebench[datagen]`"
            )
        
        
    def run(self) -> None:
        """
        This method uses DuckDB to generate in-memory tables based on the specified 
        scale factor and writes them to Parquet files. It estimates the average row 
        size in MB using a sample of the data since DuckDB only supports specifying 
        the number of rows per row group. The generated tables are written to the 
        specified target folder with optimized row group sizes.
       
        Notes
        -----
        - The method creates a sample Parquet file for each table to estimate row sizes.
        - The full table is then written as Parquet files with optimized row group sizes.
        - Temporary files and in-memory tables are cleaned up after processing.
        """
        import duckdb
        import pyarrow.parquet as pq

        # cleanup target directory
        if self.fs.exists(self.target_folder_uri):
            self.fs.rm(self.target_folder_uri, recursive=True)
        self.fs.mkdirs(self.target_folder_uri, exist_ok=True)

        with duckdb.connect() as con:
            print("Generating in-memory tables")
            con.execute(f"CALL {self.GEN_UTIL}(sf={self.scale_factor})")
            tables = [row[0] for row in con.execute("SHOW TABLES").fetchall()]
            print(f"Generated in-memory tables: {tables}")

            for table in tables:
                sample_file = posixpath.join(self.target_folder_uri, f"{table}_sample.parquet")
                full_folder_uri = posixpath.join(self.target_folder_uri, table)
                # Write a sample for row size estimation
                print(f"\nSampling {table} to evaluate row count to target {self.target_row_group_size_mb}mb row groups...")
                con.execute(f"""
                    COPY (SELECT * FROM {table} LIMIT 1000000)
                    TO '{sample_file}'
                    (FORMAT 'parquet', OVERWRITE)
                """)

                with pq.ParquetFile(sample_file) as pf:
                    rg = pf.metadata.row_group(0)
                avg_row_size = rg.total_byte_size / rg.num_rows
                #print(f"{table} sample: {rg.num_rows} rows, {rg.total_byte_size / (1024*1024):.2f} MB")
                #print(f"Avg row size: {avg_row_size:.2f} bytes")
                target_size_bytes = self.target_row_group_size_mb * 1024 * 1024
                target_rows = int(target_size_bytes / avg_row_size)
                #print(f"Target ROW_GROUP_SIZE for ~{self.target_row_group_size_mb} MB: {target_rows} rows")

                # Write full table
                print(f"Writing {table} to {full_folder_uri} with ROW_GROUP_SIZE {target_rows}...")
                con.execute(f"""
                    COPY {table} TO '{full_folder_uri}'
                    (FORMAT 'parquet', ROW_GROUP_SIZE {target_rows}, PER_THREAD_OUTPUT, OVERWRITE)
                """)

                con.execute(f"DROP TABLE {table}")

                self.fs.rm(sample_file)


from .base import BaseEngine
from  .delta_rs import DeltaRs

import os
import posixpath
from typing import Any, Optional
from importlib.metadata import version

class DuckDB(BaseEngine):
    """
    DuckDB Engine
    """
    SQLGLOT_DIALECT = "duckdb"
    SUPPORTS_ONELAKE = True
    SUPPORTS_SCHEMA_PREP = True
    SUPPORTS_MOUNT_PATH = True

    def __init__(
            self, 
            schema_or_working_directory_uri: str,
            cost_per_vcore_hour: Optional[float] = None,
            storage_options: Optional[dict[str, Any]] = None
            ):
        """
        Parameters
        ----------
        schema_or_working_directory_uri : str
            The base URI where tables are stored. This could be an arbitrary directory or
            schema path within a metastore.
        cost_per_vcore_hour : float, optional
            The cost per vCore hour for the compute runtime. If None, cost calculations are auto calculated
            where possible.
        storage_options : dict, optional
            A dictionary of storage options to pass to the engine for filesystem access. Optional as LakeBench
            will attempt to read from environment variables depeneding on the compute runtime.
        """
        
        super().__init__(schema_or_working_directory_uri, storage_options)
        import duckdb
        self.duckdb = duckdb.connect()
        self.deltars = DeltaRs()
        self.catalog_name = None
        self.schema_name = None
        if self.schema_or_working_directory_uri.startswith("abfss://"):
            self.duckdb.sql(f""" CREATE OR REPLACE SECRET onelake ( TYPE AZURE, PROVIDER ACCESS_TOKEN, ACCESS_TOKEN '{os.getenv("AZURE_STORAGE_TOKEN")}') ;""")

        self.version: str = f"{version('duckdb')} (deltalake=={version('deltalake')})"
        self.cost_per_vcore_hour = cost_per_vcore_hour or getattr(self, '_autocalc_usd_cost_per_vcore_hour', None)
    
    def _create_empty_table(self, table_name: str, ddl: str):
        if not ddl.strip().startswith("CREATE OR REPLACE TABLE"):
            ddl = ddl.replace("CREATE TABLE", "CREATE OR REPLACE TABLE")
        # Create in memory table
        self.duckdb.sql(ddl)

        arrow_df = self.duckdb.sql(f"FROM {table_name}")

        # Write empty in-memory table as Delta
        self.deltars.write_deltalake(
            table_or_uri=posixpath.join(self.schema_or_working_directory_uri, table_name),
            data=arrow_df,
            mode="overwrite",
            storage_options=self.storage_options,
        )  
        # Drop the in-memory table
        self.duckdb.sql(f"DROP TABLE IF EXISTS {table_name}")

    def load_parquet_to_delta(self, parquet_folder_uri: str, table_name: str, table_is_precreated: bool = False, context_decorator: Optional[str] = None):
        arrow_df = self.duckdb.sql(f""" FROM parquet_scan('{posixpath.join(parquet_folder_uri, '*.parquet')}') """).record_batch()
        self.deltars.write_deltalake(
            table_or_uri=posixpath.join(self.schema_or_working_directory_uri, table_name),
            data=arrow_df,
            mode="overwrite",
            storage_options=self.storage_options,
        )  

    def register_table(self, table_name: str):
        """
        Register a Delta table in DuckDB.
        """
        self.duckdb.sql(f"""
            CREATE OR REPLACE VIEW {table_name} 
            AS SELECT * FROM delta_scan('{posixpath.join(self.schema_or_working_directory_uri, table_name)}')
        """)

    def execute_sql_query(self, query: str, context_decorator: Optional[str] = None):
        """
        Execute a SQL query using DuckDB.
        """
        result = self.duckdb.sql(query).df()

    def optimize_table(self, table_name: str):
        fact_table = self.deltars.DeltaTable(
            table_uri=posixpath.join(self.schema_or_working_directory_uri, table_name),
            storage_options=self.storage_options,
        )
        fact_table.optimize.compact()

    def vacuum_table(self, table_name: str, retain_hours: int = 168, retention_check: bool = True):
        fact_table = self.deltars.DeltaTable(
            table_uri=posixpath.join(self.schema_or_working_directory_uri, table_name),
            storage_options=self.storage_options,
        )
        fact_table.vacuum(retain_hours, enforce_retention_duration=retention_check, dry_run=False)


import posixpath
import importlib.util
import fsspec
from fsspec import AbstractFileSystem
import subprocess
import threading
import math
from concurrent.futures import ThreadPoolExecutor, as_completed
from lakebench.utils.path_utils import to_unix_path
from urllib.parse import urlparse

class _TPCRsDataGenerator:
    """
    Base class for TPC Rust based data generation. PLEASE DO NOT INSTANTIATE THIS CLASS DIRECTLY. Use the TPCHDataGenerator and TPCDSDataGenerator
    subclasses instead.
    """
    GEN_UTIL = ''
    GEN_TYPE = 'tpch'
    GEN_TABLE_REGISTRY = [
        'customer', 'lineitem', 'nation', 'orders', 'part',
        'partsupp', 'region', 'supplier'
    ]
    TARGET_FILE_SIZE_MAP = [
        (10, 128), # up to 10GB -> 128MB files
        (1024, 256), # up to 1TB -> 256MB files
        (5120, 512), # up to 5TB -> 512MB files
        (10240, 1024) # up to 10TB and larger -> 1GB files
    ]
    SF1000_SIZE_GB_DICT = {
        'lineitem':  152,
        'orders': 38,
        'partsupp': 26.7,
        'part': 4,
        'customer': 7.6,
        'supplier': 0.48,
        'region': 0.00,
        'nation': 0.00
    }
    
    # Class-level lock for thread-safe printing
    _print_lock = threading.Lock()

    def __init__(self, scale_factor: int, target_folder_uri: str, target_row_group_size_mb: int = 128, compression: str = "ZSTD(1)", table_list: list = None, multithreading: bool = True) -> None:
        """
        Initialize the TPC data generator with a scale factor.

        Parameters
        ----------
        scale_factor: int
            The scale factor for the data generation.
        target_folder_uri: str
            Test data will be written to this location where tables are represented as folders containing parquet files.
        target_row_group_size_mb: int, default=128
            Desired row group size for the generated parquet files.
        compression: str, default="ZSTD(1)"
            Compression codec to use for the generated parquet files.
            Supports codecs: "UNCOMPRESSED", "SNAPPY", "GZIP(compression_level)", "BROTLI(compression_level)", "LZ4", "LZ4_RAW", "LZO", "ZSTD(compression_level)"
        """
        self.scale_factor = scale_factor
        uri_scheme = urlparse(target_folder_uri).scheme
        
        # Allow local file systems: no scheme, file://, or Windows drive letters
        cloud_schemes = {'s3', 'gs', 'gcs', 'abfs', 'abfss', 'adl', 'wasb', 'wasbs'}
        
        if uri_scheme in cloud_schemes:
            raise ValueError(f"{uri_scheme} protocol is not currently supported for TPC-RS data generation. Please use a local file system path or mount the storage location.")
        
        if compression.split('(')[0] not in ["UNCOMPRESSED", "SNAPPY", "GZIP", "BROTLI", "LZ4", "LZ4_RAW", "LZO", "ZSTD"]:
            raise ValueError(f"Unsupported compression codec: {compression}")
        
        self.fs: AbstractFileSystem = fsspec.filesystem("file")
        self.target_folder_uri = to_unix_path(target_folder_uri)
        self.target_row_group_size_mb = int(target_row_group_size_mb * 2.6) # 2.6 for uncompressed-> ZSTD(1) compression ratio
        self.compression = compression
        self.table_list = table_list
        self.multithreading = multithreading

        def get_tpcgen_path():
            import shutil
            # Try shutil.which first (most reliable)
            path = shutil.which(f"{self.GEN_TYPE}gen-cli")
            if path:
                return path

            # Fallback to user Scripts directory
            from pathlib import Path
            import sys
            user_scripts = Path.home() / "AppData" / "Roaming" / "Python" / f"Python{sys.version_info.major}{sys.version_info.minor}" / "Scripts" / "tpchgen-cli.exe"
            if user_scripts.exists():
                return str(user_scripts)

            raise ImportError(f"{self.GEN_TYPE}gen-cli is used for data generation but is not installed. Install using `%pip install {self.GEN_TYPE}gen-cli`")

        self.tpcgen_exe = get_tpcgen_path()
        
    
    def run(self) -> None:
        """
        This method uses multithreading to generate individual tables in parallel using
        a rust-based TPC data generation utility. Each table is generated with an optimal
        number of parts (based on the GEN_SF1000_FILE_COUNT_MAP) to target having files around 1GB.
        """
        
        # cleanup target directory
        def clean_dir(path: str) -> None:
            if self.fs.exists(path):
                self.fs.rm(path, recursive=True)
            self.fs.mkdirs(path, exist_ok=True)

        if self.table_list is None:
            clean_dir(self.target_folder_uri)
        else:
            for table_name in self.table_list:
                table_path = posixpath.join(self.target_folder_uri, table_name)
                clean_dir(table_path)
        
        if self.table_list is None:
            tables = self.GEN_TABLE_REGISTRY
        else:
            tables = [table for table in self.GEN_TABLE_REGISTRY if table in self.table_list]
        
        print(f"🚀 Starting parallel generation of {len(tables)} tables with multithreading...")
        print(f"📊 Scale Factor: {self.scale_factor}")
        print(f"📁 Output Directory: {self.target_folder_uri}")
        
        completed_tables = []
        failed_tables = []
        
        if self.multithreading:
            with ThreadPoolExecutor() as executor:
                future_to_table = {
                    executor.submit(self._generate_table, table_name): table_name 
                    for table_name in tables
                }

                for future in as_completed(future_to_table):
                    table_name = future_to_table[future]
                    try:
                        result = future.result()
                        if result:
                            completed_tables.append(table_name)
                            print(f"✅ {table_name} - Generation completed successfully")
                        else:
                            failed_tables.append(table_name)
                            print(f"❌ {table_name} - Generation failed")
                    except Exception as exc:
                        failed_tables.append(table_name)
                        print(f"❌ {table_name} - Generation failed with exception: {exc}")
        else:
            for table_name in tables:
                result = self._generate_table(table_name)
                if result:
                    completed_tables.append(table_name)
                    print(f"✅ {table_name} - Generation completed successfully")
                else:
                    failed_tables.append(table_name)
                    print(f"❌ {table_name} - Generation failed")
        
        print(f"\n📋 Generation Summary:")
        print(f"   ✅ Successfully generated: {len(completed_tables)} tables")
        if completed_tables:
            print(f"      Tables: {', '.join(completed_tables)}")
        
        if failed_tables:
            print(f"   ❌ Failed to generate: {len(failed_tables)} tables")
            print(f"      Tables: {', '.join(failed_tables)}")
            raise RuntimeError(f"Failed to generate {len(failed_tables)} tables: {', '.join(failed_tables)}")
        else:
            print(f"🎉 All {len(tables)} tables generated successfully!")
    
    def _generate_table(self, table_name: str) -> bool:
        """
        Generate a single table using the optimal number of parts.
        
        Parameters
        ----------
        table_name: str
            Name of the table to generate
            
        Returns
        -------
        bool
            True if generation was successful, False otherwise
        """
        def find_target_size(size: float) -> int:
            for threshold_gb, target_mb in self.TARGET_FILE_SIZE_MAP:
                if size < threshold_gb:
                    return target_mb
            return 1024

        # Scale the parts based on the scale factor and size of the table
        sf1000_size_gb = self.SF1000_SIZE_GB_DICT.get(table_name, 0)
        scale_adj_size_gb = sf1000_size_gb * (self.scale_factor / 1000.0)
        target_size_mb = find_target_size(scale_adj_size_gb)
        optimal_parts = max(round(scale_adj_size_gb * 1024 / target_size_mb), 1)
                
        print(f"🔧 {table_name} - Using {optimal_parts} parts (target file size: {target_size_mb}mb)")
        
        # ensure that 128mb target files have a single row group
        adj_row_group_target_mb = 1024 if target_size_mb == 128 else self.target_row_group_size_mb
        # Build command for individual table generation
        cmd = [
            self.tpcgen_exe,
            "--scale-factor", str(self.scale_factor),
            "--output-dir", self.target_folder_uri,
            "--parts", str(optimal_parts),
            "--format", "parquet",
            "--parquet-row-group-bytes", str(adj_row_group_target_mb * 1024 * 1024),
            "--parquet-compression", self.compression,
            "--tables", table_name 
        ]

        try:
            result = subprocess.run(cmd, capture_output=True, text=True, check=True)
            if result.stdout:
                with self._print_lock:
                    print(f"📝 {table_name} output:")
                    for line in result.stdout.strip().split('\n'):
                        if line.strip():
                            print(f"   {line}")
            return True
            
        except subprocess.CalledProcessError as e:
            with self._print_lock:
                print(f"❌ {table_name} failed:")
                if e.stdout:
                    print(f"   stdout: {e.stdout}")
                if e.stderr:
                    print(f"   stderr: {e.stderr}")
            return False
        except Exception as e:
            with self._print_lock:
                print(f"❌ {table_name} failed with exception: {e}")
            return False


import time
from datetime import datetime
from contextlib import contextmanager
from ..engines.spark import Spark

@contextmanager
def timer(phase: str = "Elapsed time", test_item: str = '', engine: str = None):
    if not hasattr(timer, "results"):
        timer.results = []

    iteration = sum(1 for result in timer.results if result[0] == phase and result[1] == test_item) + 1
    
    class TimerContext:
        def __init__(self, phase: str, test_item: str, iteration: int):
            self.execution_telemetry = {}
            self.context_decorator = f"{phase} - {test_item} [i:{iteration}]"

    timer_context = TimerContext(phase, test_item, iteration)

    if isinstance(engine, Spark):
        engine.spark.sparkContext.setJobDescription(timer_context.context_decorator)
        if engine.spark_measure_telemetry:
            engine.capture_metrics.begin()
            engine.spark.sparkContext.setLocalProperty("spark.scheduler.pool", "lakebench")

    start = time.time()
    start_datetime = datetime.now()
    success = True
    error_message = None
    error_type = None


    try:
        yield timer_context
    except Exception as e:
        success = False
        error_message = str(e)
        error_type = type(e).__name__  # Capture the error type
        print(f"Error during {phase} - {test_item}... {error_type}: {error_message}")
        
    finally:
        end = time.time()
        duration = int((end - start) * 1000)
        print(f"{phase} - {test_item}{f' [i:{iteration}]' if iteration > 1 else ''}: {(duration / 1000):.2f} seconds")
        # Set execution metadata to an empty dict if it is not set or was set to anything other than a dict
        if not isinstance(timer_context.execution_telemetry, dict):
            timer_context.execution_telemetry = {}

        if isinstance(engine, Spark):
            engine.spark.sparkContext.setJobDescription(None)
            if engine.spark_measure_telemetry:
                engine.capture_metrics.end()
                listener_metrics = engine.capture_metrics.create_stagemetrics_DF()
                listener_metrics_agg = engine.capture_metrics.aggregate_stagemetrics_DF()
                listener_metrics_dict = listener_metrics_agg.toPandas().iloc[0].to_dict()
                listener_metrics_str_dict = {k: str(v) for k, v in listener_metrics_dict.items()}
                timer_context.execution_telemetry.update(listener_metrics_str_dict)
        

        timer.results.append(
            (
                phase, 
                test_item, 
                start_datetime, 
                duration, 
                iteration, 
                success, 
                f"{error_type}: {error_message}" if error_message else '', 
                timer_context.execution_telemetry
            )
        )

def _clear_results():
    if hasattr(timer, "results"):
        timer.results = []

timer.clear_results = _clear_results


from typing import List, Optional
from ..base import BaseBenchmark
from ...utils.query_utils import transpile_and_qualify_query, get_table_name_from_ddl

from ...engines.base import BaseEngine
from ...engines.spark import Spark
from ...engines.duckdb import DuckDB
from ...engines.daft import Daft
from ...engines.polars import Polars
from ...engines.sail import Sail

import importlib.resources
import inspect
import posixpath

class _LoadAndQuery(BaseBenchmark):
    """
    Base class for benchmarks that only have a simple Load and Query phase (TPC-H, TPC-DS, ClickBench). 
    PLEASE DO NOT INSTANTIATE THIS CLASS DIRECTLY. Use the subclasses instead. 
    """
    BENCHMARK_IMPL_REGISTRY = {
        Spark: None,
        DuckDB: None,
        Daft: None,
        Polars: None,
        Sail: None,
    }
    MODE_REGISTRY = ['load', 'query', 'power_test', 'load_and_query']
    BENCHMARK_NAME = ''
    TABLE_REGISTRY = [
        'call_center', 'catalog_page', 'catalog_returns', 'catalog_sales',
        'customer', 'customer_address', 'customer_demographics', 'date_dim',
        'household_demographics', 'income_band', 'inventory', 'item',
        'promotion', 'reason', 'ship_mode', 'store', 'store_returns',
        'store_sales', 'time_dim', 'warehouse', 'web_page', 'web_returns',
        'web_sales', 'web_site'
    ]
    QUERY_REGISTRY = [
        'q1', 'q2', 'q3', 'q4', 'q5', 'q6', 'q7', 'q8', 'q9', 'q10',
        'q11', 'q12', 'q13', 'q14a', 'q14b', 'q15', 'q16', 'q17', 'q18', 'q19', 'q20',
        'q21', 'q22', 'q23a', 'q23b', 'q24a', 'q24b', 'q25', 'q26', 'q27', 'q28', 'q29', 'q30',
        'q31', 'q32', 'q33', 'q34', 'q35', 'q36', 'q37', 'q38', 'q39a', 'q39b', 'q40',
        'q41', 'q42', 'q43', 'q44', 'q45', 'q46', 'q47', 'q48', 'q49', 'q50',
        'q51', 'q52', 'q53', 'q54', 'q55', 'q56', 'q57', 'q58', 'q59', 'q60',
        'q61', 'q62', 'q63', 'q64', 'q65', 'q66', 'q67', 'q68', 'q69', 'q70',
        'q71', 'q72', 'q73', 'q74', 'q75', 'q76', 'q77', 'q78', 'q79', 'q80',
        'q81', 'q82', 'q83', 'q84', 'q85', 'q86', 'q87', 'q88', 'q89', 'q90',
        'q91', 'q92', 'q93', 'q94', 'q95', 'q96', 'q97', 'q98', 'q99'
    ]
    DDL_FILE_NAME = ''
    VERSION = ''

    def __init__(
            self, 
            engine: BaseEngine, 
            scenario_name: str,
            scale_factor: Optional[int] = None,
            query_list: Optional[List[str]] = None,
            input_parquet_folder_uri: Optional[str] = None,
            result_table_uri: Optional[str] = None,
            save_results: bool = False,
            run_id: Optional[str] = None
            ):
        self.scale_factor = scale_factor
        super().__init__(engine, scenario_name, input_parquet_folder_uri, result_table_uri, save_results, run_id)
        if query_list is not None:
            expanded_query_list = []
            for query in query_list:
                if query == '*':
                    expanded_query_list.extend(self.QUERY_REGISTRY)  # Replace '*' with all queries
                else:
                    expanded_query_list.append(query)
            query_set = set(expanded_query_list)
            if not query_set.issubset(self.QUERY_REGISTRY):
                unsupported_queries = query_set - set(self.QUERY_REGISTRY)
                raise ValueError(f"Query list contains unsupported queries: {unsupported_queries}. Supported queries: {self.QUERY_REGISTRY}.")
            self.query_list = expanded_query_list
        else:
            self.query_list = self.QUERY_REGISTRY

        for base_engine, benchmark_impl in self.BENCHMARK_IMPL_REGISTRY.items():
            if isinstance(engine, base_engine):
                self.benchmark_impl_class = benchmark_impl
                break
        else:
            raise ValueError(
                f"No benchmark implementation registered for engine type: {type(engine).__name__} "
                f"in benchmark '{self.__class__.__name__}'."
            )

        self.engine = engine
        self.scenario_name = scenario_name

        self.input_parquet_folder_uri = input_parquet_folder_uri

        self.benchmark_impl = self.benchmark_impl_class(self.engine) if self.benchmark_impl_class is not None else None

    def run(self, mode: str = 'power_test'):
        """
        Executes a specific test mode based on the provided mode string.

        Parameters
        ----------
        mode : str, optional
            The mode of the test to run. Supported modes are:
            - 'load': Executes the load test.
            - 'query': Executes the query test.
            - 'power_test': Executes the power test (default).
            - 'load_and_query': Alias for 'power_test', runs both load and query tests.

        Notes
        -----
        The `MODE_REGISTRY` attribute contains the list of supported modes.
        """
        self.mode = 'load_and_query' if mode in ('power_test', 'load_and_query') else mode

        if mode == 'load':
            self._run_load_test()
        elif mode == 'query':
            self._run_query_test()
        elif mode in ('power_test', 'load_and_query'):
            self._run_power_test()
        else:
            raise ValueError(f"Unknown mode '{mode}'. Supported modes: {self.MODE_REGISTRY}.")
    
    def _prepare_schema(self):
        """
        Prepares the database schema for the benchmark.
        This method creates the schema if it does not exist, optionally dropping it before creation.
        It then reads the DDL (Data Definition Language) file associated with the specific benchmark,
        parses the SQL statements, and executes them to set up the schema.

        Parameters
        ----------
        None

        Notes
        -----
        - The DDL file is expected to be located in the `resources.ddl` directory corresponding to the TPC benchmark variant.
        """
        self.engine.create_schema_if_not_exists(drop_before_create=True)
        self.engine.create_external_location(self.input_parquet_folder_uri)

        engine_class_name = self.engine.__class__.__name__.lower()
        parent_class_name = self.engine.__class__.__bases__[0].__name__.lower()
        benchmark_name = self.__class__.__name__.lower()
        engine_root_lib_name = self.engine.__class__.__module__.split('.')[0]
        from_dialect = self.engine.SQLGLOT_DIALECT

        try:
            # Try to load engine-specific query first
            with importlib.resources.path(
                f"{engine_root_lib_name}.benchmarks.{benchmark_name}.resources.ddl.{engine_class_name}", 
                self.DDL_FILE_NAME
            ) as ddl_path:
                with open(ddl_path, 'r') as ddl_file:
                    ddl = ddl_file.read()                
        except (ModuleNotFoundError, FileNotFoundError):
            # Try parent engine class name if engine-specific fails
            try:
                with importlib.resources.path(
                    f"lakebench.benchmarks.{benchmark_name}.resources.ddl.{parent_class_name}", 
                    self.DDL_FILE_NAME
                ) as ddl_path:
                    with open(ddl_path, 'r') as ddl_file:
                        ddl = ddl_file.read()
            except (ModuleNotFoundError, FileNotFoundError):
                # Fall back to canonical query
                with importlib.resources.path(
                    f"lakebench.benchmarks.{benchmark_name}.resources.ddl.canonical", 
                    self.DDL_FILE_NAME
                ) as ddl_path:
                    with open(ddl_path, 'r') as ddl_file:
                        ddl = ddl_file.read()
                from_dialect = 'spark'
            
        statements = [s for s in ddl.split(';') if len(s) > 7]
        for statement in statements:
            prepped_ddl = transpile_and_qualify_query(
                query=statement, 
                from_dialect=from_dialect, 
                to_dialect=self.engine.SQLGLOT_DIALECT, 
                catalog=getattr(self.engine, 'catalog_name', None),
                schema=getattr(self.engine, 'schema_name', None)
            )
            table_name = get_table_name_from_ddl(prepped_ddl)

            self.engine._create_empty_table(table_name=table_name, ddl=prepped_ddl)
            
    def _run_load_test(self):
        """
        Executes the load test by loading data from Parquet files into Delta tables 
        for all tables registered in the `TABLE_REGISTRY`. This method also measures 
        the time taken for each table load operation and records the results.

        Parameters
        ----------
        None

        Notes
        -----
        - If the engine is an instance of `Spark`, the schema is prepared before 
          loading the data.
        - The method uses a timer to measure the duration of the load operation 
          for each table.
        - Results are posted after all tables have been processed.
        """
        # set the mode if the module is being called directly
        if inspect.currentframe().f_back.f_code.co_name not in ('run', '_run_power_test'):
            self.mode = 'load'

        if self.engine.SUPPORTS_SCHEMA_PREP:
            self._prepare_schema()
        for table_name in self.TABLE_REGISTRY:
            with self.timer(phase="Load", test_item=table_name, engine=self.engine) as tc:
                if self.benchmark_impl is not None:
                    # If a specific benchmark implementation is defined, use it to load the table
                    tc.execution_telemetry = self.benchmark_impl.load_parquet_to_delta(
                        parquet_folder_uri=self.input_parquet_folder_uri,
                        table_name=table_name, 
                        table_is_precreated=True,
                        context_decorator=tc.context_decorator
                    )
                else:
                    # Otherwise, use the generic load method
                    tc.execution_telemetry = self.engine.load_parquet_to_delta(
                        parquet_folder_uri=posixpath.join(self.input_parquet_folder_uri, f"{table_name}/"), 
                        table_name=table_name,
                        table_is_precreated=True,
                        context_decorator=tc.context_decorator
                    )
        self.post_results()

    def _run_query_test(self):
        """
        Executes a series of SQL queries defined in the `query_list` attribute.
        """
        # set the mode if the module is being called directly
        if inspect.currentframe().f_back.f_code.co_name not in ('run', '_run_power_test'):
            self.mode = 'query'

        if isinstance(self.engine, (DuckDB, Daft, Polars, Sail)):
            for table_name in self.TABLE_REGISTRY:
                self.engine.register_table(table_name)
        for query_name in self.query_list:
            prepped_query = self._return_query_definition(query_name)
            with self.timer(phase="Query", test_item=query_name, engine=self.engine) as tc:
                if self.benchmark_impl is not None:
                    # If a specific benchmark implementation is defined, use it to perform the query
                    tc.execution_telemetry = self.benchmark_impl.execute_sql_query(
                        prepped_query,
                        context_decorator=tc.context_decorator
                    )
                else:
                    # Otherwise, use the generic query method
                    tc.execution_telemetry = self.engine.execute_sql_query(
                        prepped_query,
                        context_decorator=tc.context_decorator
                    )
        self.post_results()

    def _run_power_test(self):
        """
        Executes the full benchmark by running both the load and query phases.

        This method orchestrates:
        1. Load phase: Loads data into the target system.
        2. Query phase: Executes configured SQL queries to evaluate performance.
        """
        self.mode = 'load_and_query'

        self._run_load_test()
        self._run_query_test()

    def _return_query_definition(self, query_name: str) -> str:
        """
        Returns the SQL definition for a given query name.

        Parameters
        ----------
        query_name : str
            The name of the query to retrieve.

        Returns
        -------
        str
            The SQL definition for the specified query.
        """
        engine_class_name = self.engine.__class__.__name__.lower()
        parent_class_name = self.engine.__class__.__bases__[0].__name__.lower()
        benchmark_name = self.__class__.__name__.lower()
        engine_root_lib_name = self.engine.__class__.__module__.split('.')[0]
        from_dialect = self.engine.SQLGLOT_DIALECT

        try:
            # Try to load engine-specific query first
            with importlib.resources.path(
                f"{engine_root_lib_name}.benchmarks.{benchmark_name}.resources.queries.{engine_class_name}", 
                f'{query_name}.sql'
            ) as query_path:
                with open(query_path, 'r') as query_file:
                    query = query_file.read()                
        except (ModuleNotFoundError, FileNotFoundError):
            # Try parent engine class name if engine-specific fails
            try:
                with importlib.resources.path(
                    f"lakebench.benchmarks.{benchmark_name}.resources.queries.{parent_class_name}", 
                    f'{query_name}.sql'
                ) as query_path:
                    with open(query_path, 'r') as query_file:
                        query = query_file.read()
            except (ModuleNotFoundError, FileNotFoundError):
                # Fall back to canonical query
                with importlib.resources.path(
                    f"lakebench.benchmarks.{benchmark_name}.resources.queries.canonical", 
                    f'{query_name}.sql'
                ) as query_path:
                    with open(query_path, 'r') as query_file:
                        query = query_file.read()
                from_dialect = 'spark'

        prepped_query = transpile_and_qualify_query(
            query=query, 
            from_dialect=from_dialect, 
            to_dialect=self.engine.SQLGLOT_DIALECT, 
            catalog=getattr(self.engine, 'catalog_name', None),
            schema=getattr(self.engine, 'schema_name', None)
        )
        return prepped_query


from .base import BaseEngine
from .daft import Daft
from .delta_rs import DeltaRs
from .duckdb import DuckDB
from .polars import Polars
from .spark import Spark
from .fabric_spark import FabricSpark
from .synapse_spark import SynapseSpark
from .hdi_spark import HDISpark
from .sail import Sail


from .base import BaseEngine
from .delta_rs import DeltaRs

import os
import posixpath
from typing import Any, Optional
from importlib.metadata import version



class Sail(BaseEngine):
    """
    Sail Engine

    File system support: https://docs.lakesail.com/sail/main/guide/storage/
    """
    _SAIL_SERVER = None
    _SPARK = None
    SQLGLOT_DIALECT = "spark"
    SUPPORTS_ONELAKE = True
    SUPPORTS_SCHEMA_PREP = False
    SUPPORTS_MOUNT_PATH = True

    def __init__(
        self,
        schema_or_working_directory_uri: str,
        cost_per_vcore_hour: Optional[float] = None,
        storage_options: Optional[dict[str, Any]] = None
    ):
        """
        Parameters
        ----------
        schema_or_working_directory_uri : str
            The base URI where tables are stored. This could be an arbitrary directory or
            schema path within a metastore.
        cost_per_vcore_hour : float, optional
            The cost per vCore hour for the compute runtime. If None, cost calculations are auto calculated
            where possible.
        storage_options : dict, optional
            A dictionary of storage options to pass to the engine for filesystem access. Optional as LakeBench
            will attempt to read from environment variables depeneding on the compute runtime.
        """
        
        super().__init__(schema_or_working_directory_uri, storage_options)
        from pysail.spark import SparkConnectServer
        from pyspark.sql import SparkSession
        self.deltars = DeltaRs()
        self.catalog_name = None
        self.schema_name = None
        
        # Set Sail specific environment variables
        os.environ["SAIL_OPTIMIZER__ENABLE_JOIN_REORDER"] = "true"

        if Sail._SAIL_SERVER is None:
            # create server
            server = SparkConnectServer(port=50051)
            server.start(background=True)
            Sail._SAIL_SERVER = server
        self.sail_server = Sail._SAIL_SERVER

        if Sail._SPARK is None:
            sail_server_hostname, sail_server_port = self.sail_server.listening_address
            try:
                spark = SparkSession.builder.remote(
                    f"sc://{sail_server_hostname}:{sail_server_port}"
                ).getOrCreate()
                spark.conf.set("spark.sql.warehouse.dir", schema_or_working_directory_uri)
                Sail._SPARK = spark
            except ImportError as ex:
                raise RuntimeError(
                    "Python kernel restart is required after package upgrade.\nRun `import sys; sys.exit(0)` in a separate cell before initializing Sail engine."
                ) from ex
        self.spark = Sail._SPARK

        self.version: str = (
            f"""{version("pysail")} (deltalake=={version("deltalake")})"""
        )
        self.cost_per_vcore_hour = cost_per_vcore_hour or getattr(
            self, "_autocalc_usd_cost_per_vcore_hour", None
        )

    def load_parquet_to_delta(
        self,
        parquet_folder_uri: str,
        table_name: str,
        table_is_precreated: bool = False,
        context_decorator: Optional[str] = None,
    ):
        self.spark.read.parquet(parquet_folder_uri) \
            .write.format("delta") \
            .mode("overwrite") \
            .save(posixpath.join(self.schema_or_working_directory_uri, table_name))

    def register_table(self, table_name: str):
        """
        Register a Delta table as temporary view in Sail.
        """
        self.spark.read.format("delta").load(
            posixpath.join(self.schema_or_working_directory_uri, table_name)
        ).createOrReplaceTempView(table_name)

    def execute_sql_query(self, query: str, context_decorator: Optional[str] = None):
        """
        Execute a SQL query using Sail.
        """
        execute_sql = self.spark.sql(query).toPandas()

    def execute_sql_statement(self, statement: str, context_decorator: Optional[str] = None):
        """
        Execute a SQL statement.

        Parameters
        ----------
        statement : str
            The SQL statement to execute.
        context_decorator : Optional[str]
            Not used by Spark, a job description is set instead.

        """
        self.spark.sql(statement)

    def optimize_table(self, table_name: str):
        fact_table = self.deltars.DeltaTable(
            table_uri=posixpath.join(self.schema_or_working_directory_uri, table_name),
            storage_options=self.storage_options,
        )
        fact_table.optimize.compact()

    def vacuum_table(
        self, table_name: str, retain_hours: int = 168, retention_check: bool = True
    ):
        fact_table = self.deltars.DeltaTable(
            table_uri=posixpath.join(self.schema_or_working_directory_uri, table_name),
            storage_options=self.storage_options,
        )
        fact_table.vacuum(
            retain_hours, enforce_retention_duration=retention_check, dry_run=False
        )



from ....engines.duckdb import DuckDB
import posixpath
from typing import Optional

class DuckDBClickBench:
    def __init__(self, engine: DuckDB):
        
        self.engine = engine

    def load_parquet_to_delta(self, parquet_folder_uri: str, table_name: str, table_is_precreated: bool = False, context_decorator: str = None):
        """
        Loads the ClickBench parquet data into Delta format using Spark.

        Parameters
        ----------
        parquet_folder_uri : str
            Path to the source parquet files.
        """
        arrow_df = self.engine.duckdb.sql(f"""
            SELECT * REPLACE (make_date(EventDate) AS EventDate) 
            FROM parquet_scan('{posixpath.join(parquet_folder_uri, '*.parquet')}')
        """).record_batch()
        
        self.engine.deltars.write_deltalake(
            table_or_uri=posixpath.join(self.engine.schema_or_working_directory_uri, table_name),
            data=arrow_df,
            mode="append",
            storage_options=self.engine.storage_options,
        ) 

    def execute_sql_query(self, query: str, context_decorator: Optional[str] = None):
        return self.engine.execute_sql_query(query)


def abfss_to_https(abfss_path: str) -> str:
    """
    Convert an ABFSS path to an HTTPS URL.
    
    Example:
        abfss_path = "abfss://
    """
    import posixpath
    storage_account_endpoint = abfss_path.split('@')[1].split('/')[0]
    container = abfss_path.split('@')[0].split('abfss://')[1]
    file_path = abfss_path.split('@')[1].split('/')[1:]
    https_parquet_folder_path = posixpath.join('https://', storage_account_endpoint,  container, '/'.join(file_path))

    return https_parquet_folder_path

def to_unix_path(path_str) -> str:
    # Handle Windows drive letters and backslashes
    result = path_str.replace('\\', '/')
    
    # Remove Windows drive letters (C:, D:, etc.)
    if len(result) >= 2 and result[1] == ':':
        result = result[2:]
    
    # Ensure it starts with '/'
    if not result.startswith('/'):
        result = '/' + result
        
    return result


from ....engines.sail import Sail

import posixpath

class SailELTBench:
    def __init__(self, engine: Sail):
        
        import numpy as np
        self.np = np
        self.engine = engine

    def create_total_sales_fact(self):
        for table in ['store_sales', 'date_dim', 'store', 'item', 'customer']:
            self.engine.register_table(table)

        df = self.engine.spark.sql("""
            SELECT 
                s.s_store_id,
                i.i_item_id,
                c.c_customer_id,
                d.d_date AS sale_date,
                SUM(ss.ss_quantity) AS total_quantity,
                SUM(ss.ss_net_paid) AS total_net_paid,
                SUM(ss.ss_net_profit) AS total_net_profit
            FROM 
                store_sales ss
            JOIN 
                date_dim d ON ss.ss_sold_date_sk = d.d_date_sk
            JOIN 
                store s ON ss.ss_store_sk = s.s_store_sk
            JOIN 
                item i ON ss.ss_item_sk = i.i_item_sk
            JOIN 
                customer c ON ss.ss_customer_sk = c.c_customer_sk
            WHERE 
                d.d_year = 2001
            GROUP BY 
                s.s_store_id, i.i_item_id, c.c_customer_id, d.d_date
            ORDER BY 
                s.s_store_id, d.d_date;
        """)

        df.write.format("delta").mode("overwrite").save(posixpath.join(self.engine.schema_or_working_directory_uri, 'total_sales_fact'))

    def merge_percent_into_total_sales_fact(self, percent: float):
        seed = self.np.random.randint(1, high=1000, size=None, dtype=int)
        modulo = int(1 / percent)

        sampled_fact_data = self.engine.spark.sql(f"""
            SELECT 
                s.s_store_id, 
                i.i_item_id, 
                CASE 
                    WHEN rand() > 0.5 THEN 
                        CONCAT('NEW_', CAST(new_uid_val AS STRING))
                    ELSE 
                        c.c_customer_id
                END AS c_customer_id,
                d.d_date AS sale_date,
                ss.ss_quantity + FLOOR(rand() * 5 + 1) AS total_quantity,
                ss.ss_net_paid + rand() * 50 + 5 AS total_net_paid,
                ss.ss_net_profit + rand() * 20 + 1 AS total_net_profit
            FROM (
                SELECT *, 
                       ss_customer_sk + ss_sold_date_sk + {seed} AS new_uid_val
                FROM store_sales
                WHERE MOD(ss_customer_sk + ss_sold_date_sk + {seed}, {modulo}) = 0
            ) ss
            JOIN date_dim d 
              ON ss.ss_sold_date_sk = d.d_date_sk
            JOIN store s 
              ON ss.ss_store_sk = s.s_store_sk
            JOIN item i 
              ON ss.ss_item_sk = i.i_item_sk
            JOIN customer c 
              ON ss.ss_customer_sk = c.c_customer_sk
            """).toArrow()

        fact_table = self.engine.deltars.DeltaTable(
            table_uri=posixpath.join(self.engine.schema_or_working_directory_uri, 'total_sales_fact'),
            storage_options=self.engine.storage_options,
        )

        fact_table.merge(
                source=sampled_fact_data,
                predicate="""
                target.s_store_id = source.s_store_id AND 
                target.i_item_id = source.i_item_id AND 
                target.c_customer_id = source.c_customer_id AND 
                target.sale_date = source.sale_date
                """,
                source_alias="source",
                target_alias="target"
            ) \
            .when_matched_update(
                {
                    "total_quantity": "target.total_quantity + source.total_quantity",
                    "total_net_paid": "target.total_net_paid + source.total_net_paid",
                    "total_net_profit": "target.total_net_profit + source.total_net_profit",
                }
            ) \
            .when_not_matched_insert(
                {
                    "s_store_id": "source.s_store_id",
                    "i_item_id": "source.i_item_id",
                    "c_customer_id": "source.c_customer_id",
                    "sale_date": "source.sale_date",
                    "total_quantity": "source.total_quantity",
                    "total_net_paid": "source.total_net_paid",
                    "total_net_profit": "source.total_net_profit",
                }
            ) \
            .execute()
        
    def query_total_sales_fact(self):
        self.engine.register_table('total_sales_fact')
        df = self.engine.spark.sql(f"""
                            select sum(total_net_profit), year(sale_date) 
                            from total_sales_fact group by year(sale_date)
                            """)
        result = df.collect()


import posixpath
from typing import Optional


class ClickBenchDataGenerator:

    def __init__(self, target_mount_folder_uri: str = None, partitioned_files: bool = True):
        """
        Initialize the ClickBench data generator. Technically, this just downloads the ClickBench data from the ClickHouse datasets repository.

        :param partitioned_files: If True, the downloaded data will be 100 partitioned files, otherwise it is one massive file. Use partitioned files for better download performance. 
        """
        self.target_mount_folder_path = target_mount_folder_uri
        self.partitioned_files = partitioned_files


    def run(self):
        """
        Download ClickBench Parquet files to the target folder.

        This method fetches data from the ClickHouse datasets repository and writes
        it to `self.target_mount_folder_path`. If `self.partitioned_files` is True,
        it will download 100 partitioned files using multithreading (hits_0.parquet … hits_99.parquet);
        otherwise it downloads a single file (hits.parquet).

        Notes
        -----
        - Use self.partitioned_files = True for better download performance.
        - Both the partitioned and non-partition parquet data is made up of ~ 180MB row groups so it should
        have little impact on performance for most systems.
        """

        if self.partitioned_files:
            from concurrent.futures import ThreadPoolExecutor
            with ThreadPoolExecutor() as executor:
                executor.map(self.__download_parquet, range(100))
        else:
            self.__download_parquet(None)

    def __download_parquet(self, file_index: Optional[int] = None):
        file_name = f"hits_{file_index}.parquet" if file_index is not None else "hits.parquet"
        source_folder = 'athena_partitioned' if file_index is not None else 'athena'

        import urllib.request
        url = f"https://datasets.clickhouse.com/hits_compatible/{source_folder}/{file_name}"
        local_path = posixpath.join(self.target_mount_folder_path, file_name)

        headers = {'User-Agent': 'Mozilla/5.0'}
        req = urllib.request.Request(url, headers=headers)

        try:
            with urllib.request.urlopen(req) as response, open(local_path, 'wb') as out_file:
                out_file.write(response.read())
            print(f"Downloaded {file_name}")
        except Exception as e:
            print(f"Failed to download {file_name}: {e}")


from .spark import Spark
from typing import Optional
from decimal import Decimal

class SynapseSpark(Spark):
    """
    Synapse Spark Engine
    """

    def __init__(
            self,
            schema_name: str,
            schema_uri: Optional[str] = None,
            spark_measure_telemetry: bool = False,
            cost_per_vcore_hour: Optional[float] = None
            ):
        """
        Parameters
        ----------
        schema_name : str
            The name of the schema (database) to use within the catalog.
        schema_uri : str, optional
            The URI of the schema (database) to use within the catalog.
        spark_measure_telemetry : bool, default False
            Whether to enable sparkmeasure telemetry for performance measurement.
        cost_per_vcore_hour : float, optional
            The cost per vCore hour for the Spark cluster. If None, cost calculations are auto calculated
            where possible.
        """

        super().__init__(
            catalog_name=None, 
            schema_name=schema_name, 
            schema_uri=schema_uri,
            spark_measure_telemetry=spark_measure_telemetry,
            cost_per_vcore_hour=cost_per_vcore_hour,
            compute_stats_all_cols=False
            )        

        if not self.runtime != 'synapse':
            raise RuntimeError("This engine is only supports Synapse Spark Pools.")
        self.version: str = f"{self.spark.sparkContext.version} (vhd_name=={self.spark.conf.get('spark.synapse.vhd.name')})"
        region = self.spark.conf.get('spark.cluster.region')
        self.cost_per_vcore_hour = cost_per_vcore_hour if cost_per_vcore_hour is not None else self._get_vm_retail_rate(region=region, sku='vCore')
        self.cost_per_hour = self.get_total_cores() * self.cost_per_vcore_hour

        self.extended_engine_metadata.update({
            'spark_history_url': self.spark_configs['spark.tracking.webUrl'],
            'cost_per_hour': Decimal(self.cost_per_hour).quantize(Decimal('0.0000')),
            'compute_region': region
        })

        spark_configs_to_log = {k: v for k, v in self.spark_configs.items() if k in [
            'spark.microsoft.delta.optimizeWrite.enabled',
            'spark.microsoft.delta.optimizeWrite.binSize',
            'spark.synapse.vegas.useCache',
            'spark.synapse.vegas.cacheSize',
            'spark.synapse.vhd.name',
            'spark.synapse.vhd.id',
            'spark.app.id',
            'spark.cluster.name'
        ]}

        self.extended_engine_metadata.update(spark_configs_to_log)

    def _get_vm_retail_rate(self, region: str, sku: str, spot: bool = False) -> float:
        import requests
        query = f"armRegionName eq '{region}' and serviceName eq 'Azure Synapse Analytics' and productName eq 'Azure Synapse Analytics Serverless Apache Spark Pool - Memory Optimized'"
        api_url = "https://prices.azure.com/api/retail/prices?"
        return requests.get(api_url, params={'$filter': query}).json()['Items'][0]['retailPrice']
    


from .base import BaseEngine
import os
from typing import Optional
import posixpath
import tenacity

class Spark(BaseEngine):
    """
    Generic Spark Engine

    Attributes
    ----------
    SQLGLOT_DIALECT : str, optional
        Specifies the SQL dialect to be used by the engine when SQL transpiling
        is required. Default is None.
    SUPPORTS_ONELAKE : bool
        Indicates if the engine supports OneLake URIs (e.g., abfss://workspace@onelake.dfs.fabric.microsoft.com/...)
    SUPPORTS_SCHEMA_PREP : bool
        Indicates if the engine supports schema preparation (creation of empty table with defined schema)
    SUPPORTS_MOUNT_PATH : bool
        Indicates if the engine supports mount URIs (e.g., /mnt/...)

    Methods
    -------
    get_total_cores()
        Returns the total number of CPU cores available on the system.
    get_compute_size()
        Returns a formatted string with the compute size.
    append_array_to_delta(abfss_path: str, array: list)
        Appends a list of data to a Delta table at the specified path.
    """
    SQLGLOT_DIALECT = "spark"
    SUPPORTS_MOUNT_PATH = True
    SUPPORTS_ONELAKE = True
    SUPPORTS_SCHEMA_PREP = True
    

    def __init__(
            self,
            schema_name: str,
            catalog_name: Optional[str] = None,
            schema_uri: Optional[str] = None,
            spark_measure_telemetry: bool = False,
            cost_per_vcore_hour: Optional[float] = None,
            compute_stats_all_cols: bool = False
            ):
        """
        Parameters
        ----------
        schema_name : str
            The name of the schema (database) to use within the catalog.
        catalog_name : str, optional
            The name of the catalog (metastore) to use. If None, the default catalog is used.
        schema_uri : str, optional
            The URI of the schema (database) to use within the catalog.
        spark_measure_telemetry : bool, default False
            Whether to enable sparkmeasure telemetry for performance measurement.
        cost_per_vcore_hour : float, optional
            The cost per vCore hour for the Spark cluster. If None, cost calculations are auto calculated
            where possible.
        compute_stats_all_cols : bool, default False
            Whether to compute statistics for all columns after each table is loaded.
        """
        super().__init__(schema_or_working_directory_uri=schema_uri)
        from pyspark.sql import SparkSession
        import pyspark.sql.functions as sf
        self.sf = sf

        self.spark = SparkSession.builder
        if self.runtime == "local_unknown":
            warehouse_dir = posixpath.dirname(schema_uri.rstrip('/').rstrip('\\'))
            self.spark = (
                self.spark
                    .master("local[*]")
                    .config("spark.sql.warehouse.dir", warehouse_dir)
                    .config("spark.driver.host", "localhost")
                    .config("spark.driver.bindAddress", "localhost")
                    .config("spark.ui.enabled", "false")
                    .config("spark.sql.extensions", "io.delta.sql.DeltaSparkSessionExtension")
                    .config("spark.sql.catalog.spark_catalog", "org.apache.spark.sql.delta.catalog.DeltaCatalog")
                    .config("spark.jars.packages", "io.delta:delta-spark_2.12:3.2.0")
                    .config("spark.sql.catalogImplementation", "hive")
            )
            if self.operating_system == "windows":
                # Windows-specific configurations to avoid native IO issues
                self.spark = (
                    self.spark
                        .config("spark.hadoop.io.native.lib.available", "false")
                        .config("spark.hadoop.fs.file.impl.disable.cache", "true")
                )

        self.spark = self.spark.getOrCreate()

        self.schema_uri = schema_uri
        if spark_measure_telemetry:
            try:
                from sparkmeasure import StageMetrics
                self.capture_metrics = StageMetrics(self.spark)
            except ModuleNotFoundError:
                raise ModuleNotFoundError("`sparkmeasure` is not installed, either disable the `spark_measure_telemetry` flag, run `%pip install sparkmeasure==0.24.0`, or install LakeBench with the sparkmeasure option: `%pip install lakebench[sparkmeasure]`.")
        self.spark_measure_telemetry = spark_measure_telemetry

        self.version: str = self.spark.sparkContext.version

        self.catalog_name = catalog_name if self.runtime != "local_unknown" else None
        self.schema_name = schema_name
        self.full_catalog_schema_reference : str = f"`{self.catalog_name}`.`{self.schema_name}`" if catalog_name else f"`{self.schema_name}`"
        self.cost_per_vcore_hour = cost_per_vcore_hour
        self.spark_configs = self.__get_spark_session_configs()
        self.extended_engine_metadata.update({
            'parquet.block.size': self.spark.sparkContext._jsc.hadoopConfiguration().get("parquet.block.size") or '',
        })
        spark_configs_to_log = {k: v for k, v in self.spark_configs.items() if k in [
            'spark.executor.memory',
            'spark.databricks.delta.optimizeWrite.enabled',
            'spark.databricks.delta.optimizeWrite.binSize',
            'spark.sql.autoBroadcastJoinThreshold',
            'spark.sql.sources.parallelPartitionDiscovery.parallelism',
            'spark.sql.cbo.enabled',
            'spark.sql.shuffle.partitions',
            'spark.task.cpus',
            'spark.sql.parquet.compression.codec'
        ]}

        self.extended_engine_metadata.update(spark_configs_to_log)

        self.compute_stats_all_cols = compute_stats_all_cols
        self.run_analyze_after_load = self.compute_stats_all_cols

    def __get_spark_session_configs(self) -> dict:
        """
        Extract all Spark session configuration parameters.

        Returns
        -------
        dict
            Dictionary containing all Spark configuration key-value pairs.
        """
        scala_map = self.spark.conf._jconf.getAll()
        spark_conf_dict = {}
 
        iterator = scala_map.iterator()
        while iterator.hasNext():
            entry = iterator.next()
            key = entry._1()
            value = entry._2()
            spark_conf_dict[key] = value
        return spark_conf_dict
    
    # Use tenacity to retry on NativeIO error common in spark running on local Windows
    @tenacity.retry(
        retry=tenacity.retry_if_exception(
            lambda e: "java.lang.UnsatisfiedLinkError" in str(e) and 
                     "NativeIO$POSIX.stat" in str(e)
        ),
        stop=tenacity.stop_after_attempt(2)
    )
    def create_schema_if_not_exists(self, drop_before_create: bool = True):
        """
        Prepare an empty schema in the lakehouse.

        Parameters
        ----------
        drop_before_create : bool, default True
            Whether to drop the schema before creating it if it already exists.

        Notes
        -----
        Uses tenacity retry decorator to handle NativeIO errors common in Spark
        running on local Windows environments.
        """
        location_str = f"LOCATION '{self.schema_uri}'" if self.schema_uri is not None else ''

        if drop_before_create:
            self.spark.sql(f"DROP SCHEMA IF EXISTS {self.full_catalog_schema_reference} CASCADE")
        self.spark.sql(f"CREATE SCHEMA IF NOT EXISTS {self.full_catalog_schema_reference} {location_str}")
        self.spark.sql(f"USE {self.full_catalog_schema_reference}")

    def _create_empty_table(self, table_name: Optional[str], ddl: str):
        """
        Create an empty table using the provided DDL statement.

        Parameters
        ----------
        table_name : str, optional
            The name of the table to create (for compatibility, not used).
        ddl : str
            The DDL statement to execute for table creation.

        Notes
        -----
        Automatically adds 'USING delta' clause if no storage format is specified.
        """
        # Explicitly set the table type to Delta if not already specified
        if 'using ' not in ddl.lower():
            # Find the closing parenthesis of the column definitions
            closing_paren_index = ddl.rfind(")")
            if closing_paren_index != -1:
                # Insert 'USING delta' after the closing parenthesis
                ddl = (
                    ddl[:closing_paren_index + 1]
                    + " using delta"
                    + ddl[closing_paren_index + 1:]
                )

        self.execute_sql_statement(ddl)

    def _convert_generic_to_specific_schema(self, generic_schema: list):
        """
        Convert a generic schema to a specific Spark schema.
        """
        from pyspark.sql.types import StructType, StructField, StringType, IntegerType, FloatType, DoubleType, BooleanType, TimestampType, MapType, ByteType, ShortType, LongType, DecimalType
        type_mapping = {
            'STRING': StringType(),
            'TIMESTAMP': TimestampType(),
            'TINYINT': ByteType(),
            'SMALLINT': ShortType(),
            'INT': IntegerType(),
            'BIGINT': LongType(),
            'FLOAT': FloatType(),
            'DOUBLE': DoubleType(),
            'DECIMAL(18,10)': DecimalType(18,10),  # Spark does not have a specific Decimal type, using DoubleType
            'BOOLEAN': BooleanType(),
            'MAP<STRING, STRING>': MapType(StringType(), StringType())
        }
        return StructType([StructField(name, type_mapping[data_type], True) for name, data_type in generic_schema])

    def _append_results_to_delta(self, table_uri: str, results: list, generic_schema: list):
        """
        Append an array to a Delta table.
        """
        import pyspark.sql.functions as sf
        schema = self._convert_generic_to_specific_schema(generic_schema)
        # Use default order of columns in dictionary
        columns = list(results[0].keys())
        df = self.spark.createDataFrame(results, schema=schema).select(*columns)
        df.write.format("delta") \
            .option("mergeSchema", "true") \
            .option("delta.enableDeletionVectors", "false") \
            .option("delta.autoOptimize.autoCompact", "true") \
            .option("delta.autoOptimize.optimizeWrite", "true") \
            .mode("append") \
            .save(table_uri)

    def get_total_cores(self) -> int:
        """
        Returns the total number of CPU cores available in the Spark cluster.
        
        Assumes that the driver and workers nodes are all the same VM size.
        """
        cores = int(len(set(executor.host() for executor in self.spark.sparkContext._jsc.sc().statusTracker().getExecutorInfos())) * os.cpu_count())
        return cores
        
    def get_compute_size(self) -> str:
        """
        Returns a formatted string with the compute size.
        
        Assumes that the driver and workers nodes are all the same VM size.
        """        
        sc_conf_dict = {key: value for key, value in self.spark.sparkContext.getConf().getAll()}
        executor_count = self.spark.sparkContext._jsc.sc().getExecutorMemoryStatus().size() - 1
        executor_cores = int(sc_conf_dict.get('spark.executor.cores', os.cpu_count()))
        vm_host_count = len(set(executor.host() for executor in self.spark.sparkContext._jsc.sc().statusTracker().getExecutorInfos()))
        worker_count = vm_host_count - 1
        worker_cores = os.cpu_count()
        as_min_workers = sc_conf_dict['spark.dynamicAllocation.initialExecutors'] if sc_conf_dict.get('spark.autoscale.executorResourceInfoTag.enabled', 'false') == 'true' else None
        as_max_workers = sc_conf_dict['spark.dynamicAllocation.maxExecutors'] if sc_conf_dict.get('spark.autoscale.executorResourceInfoTag.enabled', 'false') == 'true' else None
        as_enabled = True if as_min_workers != as_max_workers and sc_conf_dict.get('spark.dynamicAllocation.minExecutors', None) != sc_conf_dict.get('spark.dynamicAllocation.maxExecutors', None) else False
        type = "SingleNode" if vm_host_count == 1 and not as_enabled else 'MultiNode'
        workers_word = 'Workers' if worker_count > 1 or (as_max_workers is not None and int(as_max_workers) > 1)  else 'Worker'
        executors_per_worker = int(executor_count / worker_count) if worker_count > 0 else 1
        executors_word = 'Executors' if executors_per_worker > 1 else 'Executor'
        executor_str = f"({executors_per_worker} x {executor_cores}vCore {executors_word}{' ea.' if type != 'SingleNode' else ''})"

        if type == 'SingleNode':
            cluster_config = f"{worker_cores}vCore {type} {executor_str}"
        elif as_enabled:
            cluster_config = f"{as_min_workers}-{as_max_workers} x {worker_cores}vCore {workers_word} {executor_str}"
        else:
            cluster_config = f"{worker_count} x {worker_cores}vCore {workers_word} {executor_str}"

        return cluster_config
    
    def load_parquet_to_delta(self, parquet_folder_uri: str, table_name: str, table_is_precreated: bool = False, context_decorator: Optional[str] = None):
        df = self.spark.read.parquet(parquet_folder_uri)
        if table_is_precreated:
            df.write.insertInto(table_name, overwrite=True)
        else:
            df.write.format('delta').mode("append").saveAsTable(table_name)

        if self.run_analyze_after_load:
            self.spark.sql(f"ANALYZE TABLE {table_name} COMPUTE STATISTICS FOR ALL COLUMNS;")    

    def execute_sql_query(self, query: str, context_decorator: Optional[str] = None):
        execute_sql = self.spark.sql(query).collect()
    
    def execute_sql_statement(self, statement: str, context_decorator: Optional[str] = None):
        """
        Execute a SQL statement.

        Parameters
        ----------
        statement : str
            The SQL statement to execute.
        context_decorator : Optional[str]
            Not used by Spark, a job description is set instead.

        """
        self.spark.sql(statement)

    def optimize_table(self, table_name: str):
        self.spark.sql(f"OPTIMIZE {self.full_catalog_schema_reference}.{table_name}")

    def vacuum_table(self, table_name: str, retain_hours: int = 168, retention_check: bool = True):
        self.spark.conf.set("spark.databricks.delta.retentionDurationCheck.enabled", retention_check)
        self.spark.sql(f"VACUUM {self.full_catalog_schema_reference}.{table_name} RETAIN {retain_hours} HOURS")



from .base import BaseEngine
from .delta_rs import DeltaRs

import posixpath
from typing import Any, Optional
from importlib.metadata import version

class Polars(BaseEngine):
    """
    Polars Engine
    """
    SQLGLOT_DIALECT = "duckdb"
    SUPPORTS_ONELAKE = True
    SUPPORTS_SCHEMA_PREP = False
    SUPPORTS_MOUNT_PATH = True

    def __init__(
            self, 
            schema_or_working_directory_uri: str,
            cost_per_vcore_hour: Optional[float] = None,
            storage_options: Optional[dict[str, Any]] = None
            ):
        """
        Parameters
        ----------
        schema_or_working_directory_uri : str
            The base URI where tables are stored. This could be an arbitrary directory or
            schema path within a metastore.
        cost_per_vcore_hour : float, optional
            The cost per vCore hour for the compute runtime. If None, cost calculations are auto calculated
            where possible.
        storage_options : dict, optional
            A dictionary of storage options to pass to the engine for filesystem access. Optional as LakeBench
            will attempt to read from environment variables depeneding on the compute runtime.
        """
        
        super().__init__(schema_or_working_directory_uri, storage_options)
        import polars as pl
        self.pl = pl
        self.deltars = DeltaRs()
        self.catalog_name = None
        self.schema_name = None
        self.sql = pl.SQLContext()
        self.version: str = f"{version('polars')} (deltalake=={version('deltalake')})"
        self.cost_per_vcore_hour = cost_per_vcore_hour or getattr(self, '_autocalc_usd_cost_per_vcore_hour', None)

    def load_parquet_to_delta(self, parquet_folder_uri: str, table_name: str, table_is_precreated: bool = False, context_decorator: Optional[str] = None):
        table_df = self.pl.scan_parquet(
            posixpath.join(parquet_folder_uri, '*.parquet'), 
            storage_options=self.storage_options
        )
        table_df.collect(engine='streaming').write_delta(
            posixpath.join(self.schema_or_working_directory_uri, table_name), 
            mode="overwrite", 
            storage_options=self.storage_options
        )

    def register_table(self, table_name: str):
        """
        Register a Delta table LazyFrame in Polars.
        """
        df = self.pl.scan_delta(
            posixpath.join(self.schema_or_working_directory_uri, table_name), 
            storage_options=self.storage_options
        )
        self.sql.register(table_name, df)

    def execute_sql_query(self, query: str, context_decorator: Optional[str] = None):
        """
        Execute a SQL query using Polars.
        """
        result = self.sql.execute(query).collect(engine='streaming')

    def optimize_table(self, table_name: str):
        fact_table = self.deltars.DeltaTable(
            table_uri=posixpath.join(self.schema_or_working_directory_uri, table_name),
            storage_options=self.storage_options,
        )
        fact_table.optimize.compact()

    def vacuum_table(self, table_name: str, retain_hours: int = 168, retention_check: bool = True):
        fact_table = self.deltars.DeltaTable(
            table_uri=posixpath.join(self.schema_or_working_directory_uri, table_name),
            storage_options=self.storage_options,
        )
        fact_table.vacuum(retain_hours, enforce_retention_duration=retention_check, dry_run=False)


from typing import Optional, List
from .._load_and_query import _LoadAndQuery

from ...engines.base import BaseEngine
from ...engines.spark import Spark
from ...engines.duckdb import DuckDB
from ...engines.daft import Daft
from ...engines.polars import Polars
from ...engines.sail import Sail

from .engine_impl.spark import SparkClickBench
from .engine_impl.duckdb import DuckDBClickBench
from .engine_impl.sail import SailClickBench

class ClickBench(_LoadAndQuery):
    """
    Class for running the ClickBench benchmark.

    This class provides functionality for running the ClickBench benchmark, including loading data,
    executing queries, and performing power tests. Supported engines are listed in the 
    `self.BENCHMARK_IMPL_REGISTRY` constant.

    Parameters
    ----------
    engine : object
        The engine to use for executing the benchmark.
    scenario_name : str
        The name of the benchmark scenario.
    query_list : list of str, optional
        List of queries to execute. Use '*' for all queries. If not specified, all queries will be run.
    input_parquet_folder_uri : str, optional
        Path to the input parquet files.
    result_table_uri : str, optional
        Table URI where results will be saved. Must be specified if `save_results` is True.
    save_results : bool
        Whether to save the benchmark results. Results can also be accessed via the `self.results` 
        attribute after running the benchmark.

    Methods
    -------
    run(mode='power_test')
        Runs the benchmark in the specified mode.
        Supported modes are:
            - 'load': Sequentially executes loading the 24 tables.
            - 'query': Sequentially executes the 99 queries.
            - 'power_test': Executes the load test followed by the query test.
    _run_load_test()
        Loads the data for the benchmark.
    _run_query_test()
        Executes the queries for the benchmark.
    _run_power_test()
        Runs both the load and query tests.
    """
    BENCHMARK_IMPL_REGISTRY = {
        Spark: SparkClickBench,
        DuckDB: DuckDBClickBench,
        Sail: SparkClickBench
    }
    BENCHMARK_NAME = 'ClickBench'
    TABLE_REGISTRY = [
        'hits'
    ]
    QUERY_REGISTRY = [
        'q1', 'q2', 'q3', 'q4', 'q5', 'q6', 'q7', 'q8', 'q9', 'q10',
        'q11', 'q12', 'q13', 'q14', 'q15', 'q16', 'q17', 'q18', 'q19', 'q20',
        'q21', 'q22', 'q23', 'q24', 'q25', 'q26', 'q27', 'q28', 'q29', 'q30',
        'q31', 'q32', 'q33', 'q34', 'q35', 'q36', 'q37', 'q38', 'q39', 'q40',
        'q41', 'q42', 'q43'
    ]
    DDL_FILE_NAME = 'ddl.sql'
    VERSION = 'UNKNOWN'

    def __init__(
            self, 
            engine: BaseEngine, 
            scenario_name: str,
            query_list: Optional[List[str]] = None,
            input_parquet_folder_uri: Optional[str] = None,
            result_table_uri: Optional[str] = None,
            save_results: bool = False
        ):
        super().__init__(
            engine=engine, 
            scenario_name=scenario_name,
            scale_factor=None,
            query_list=query_list,
            input_parquet_folder_uri=input_parquet_folder_uri,
            result_table_uri=result_table_uri,
            save_results=save_results
        )


from .spark import Spark
from typing import Optional

class HDISpark(Spark):
    """
    HDInsight Spark Engine
    """

    def __init__(
            self,
            schema_name: str,
            spark_measure_telemetry: bool = False,
            cost_per_vcore_hour: Optional[float] = None
            ):
        """
        Parameters
        ----------
        schema_name : str
            The name of the schema (database) to use within the catalog.
        spark_measure_telemetry : bool, default False
            Whether to enable sparkmeasure telemetry for performance measurement.
        cost_per_vcore_hour : float, optional
            The cost per vCore hour for the Spark cluster. If None, cost calculations are auto calculated
            where possible.
        """

        super().__init__(
            catalog_name=None, 
            schema_name=schema_name, 
            spark_measure_telemetry=spark_measure_telemetry,
            cost_per_vcore_hour=cost_per_vcore_hour,
            compute_stats_all_cols=False
            )
